import logging
import re
from typing import Optional

from starlette.requests import Request

from . import exceptions, glove
from .settings import BaseSettings
from .utils import get_ip

logger = logging.getLogger('foxglove.recaptcha')
TESTING_SECRET = '6LeIxAcTAAAAAGG-vFI1TnRWxMZNFuojJ4WifJWe'


async def check_recaptcha(request: Request, recaptcha_token: Optional[str], *, error_headers=None) -> None:
    client_ip = get_ip(request)

    if not recaptcha_token:
        logger.warning('recaptcha token not provided, path="%s" ip=%s', request.url.path, client_ip)
        raise exceptions.HttpBadRequest('No recaptcha value', headers=error_headers)

    settings: BaseSettings = glove.settings

    post_data = {'secret': settings.recaptcha_secret, 'response': recaptcha_token, 'remoteip': client_ip}
    r = await glove.http.post(settings.recaptcha_url, data=post_data)
    r.raise_for_status()
    data = r.json()

    request_host = request.url.hostname
    # using the origin here if available instead of host avoids problems when requests are proxied e.g. with netlify
    if origin := request.headers.get('origin'):
        request_host = re.sub('^https?://', '', origin)

    if data['success']:
        hostname = data['hostname']
        if hostname == request_host:
            logger.info('recaptcha success')
            return
        elif settings.dev_mode and settings.recaptcha_secret == TESTING_SECRET and hostname == 'testkey.google.com':
            logger.info('recaptcha test key success')
            return

    logger.warning(
        'recaptcha failure, path=%s request_host=%s ip=%s response=%s',
        request.url.path,
        request_host,
        client_ip,
        r.text,
        extra={'data': {'recaptcha_response': data, 'recaptcha_token': recaptcha_token}},
    )
    raise exceptions.HttpBadRequest('Invalid recaptcha value', headers=error_headers)


class RecaptchaDepends:
    __slots__ = ('request',)

    def __init__(self, request: Request):
        self.request = request

    async def __call__(self, recaptcha_token: Optional[str], *, error_headers=None) -> None:
        return await check_recaptcha(self.request, recaptcha_token, error_headers=error_headers)



import asyncio
import os
import secrets
from functools import wraps
from time import time
from typing import Optional, Tuple, Union

from starlette.responses import Response
from starlette.templating import Jinja2Templates as _Jinja2Templates, _TemplateResponse

from .main import glove

try:
    import jinja2
except ImportError:  # pragma: nocover
    jinja2 = None  # type: ignore

__all__ = 'FoxgloveTemplates', 'FoxgloveTestTemplates'

reload_sha = 'fbc87301a2470263e3ba45e56c7089f286a84a4e'
reload_snippet = f'<script src="https://rawcdn.githack.com/samuelcolvin/foxglove/{reload_sha}/reload.js"></script>'
static_version = (os.getenv('HEROKU_SLUG_COMMIT') or secrets.token_hex(4))[:7]


class FoxgloveTemplates(_Jinja2Templates):
    def __init__(self, directory: Optional[str] = None):
        super().__init__(glove.settings.template_dir if directory is None else directory)

    def get_env(self, directory: str) -> 'jinja2.Environment':
        @jinja2.contextfunction
        def prompt_reload(context: dict) -> str:
            if context['dev_mode']:
                return jinja2.Markup(reload_snippet)
            else:
                return ''

        @jinja2.contextfunction
        def static_url(context: dict, path: str) -> str:
            request = context['request']
            try:
                static_route_name = context.get('static_route_name', 'static')
                url = request.url_for(static_route_name, path=path)
                if context['dev_mode']:
                    url += f'?t={time() * 1000:0.0f}'
                else:
                    url += f'?v={static_version}'
                return url
            except KeyError:
                return f'/static/{path}'

        env = super().get_env(directory)
        env.globals.update(prompt_reload=prompt_reload, static_url=static_url, dev_mode=glove.settings.dev_mode)
        return env

    def render(self, template_name: str):
        def view_decorator(view):
            if asyncio.iscoroutinefunction(view):

                @wraps(view)
                async def view_wrapper(request, *args, **kwargs):
                    return self._return_template(request, template_name, await view(request, *args, **kwargs))

            else:

                @wraps(view)
                def view_wrapper(request, *args, **kwargs):
                    return self._return_template(request, template_name, view(request, *args, **kwargs))

            return view_wrapper

        return view_decorator

    def _return_template(self, request, template_name: str, r: Union[Optional[dict], Tuple[int, Optional[dict]]]):
        if isinstance(r, (tuple, list)):
            status_code, context = r
        else:
            status_code, context = 200, r
        if context is None:
            context = {}
        context['request'] = request
        return self.TemplateResponse(template_name, context, status_code=status_code)


class FoxgloveTestTemplates(FoxgloveTemplates):
    # from here on is just a hacky workaround for https://github.com/encode/starlette/issues/472
    def TemplateResponse(
        self,
        name: str,
        context: dict,
        status_code: int = 200,
        headers: dict = None,
        media_type: str = None,
        background=None,
    ) -> _TemplateResponse:
        if 'request' not in context:
            raise ValueError('context must include a "request" key')
        template = self.get_template(name)
        return CustomTemplateResponse(
            template,
            context,
            status_code=status_code,
            headers=headers,
            media_type=media_type,
            background=background,
        )


class CustomTemplateResponse(_TemplateResponse):
    async def __call__(self, scope, receive, send) -> None:
        # context sending removed
        await Response.__call__(self, scope, receive, send)



from typing import TYPE_CHECKING, Any, Dict, Optional

from starlette.testclient import TestClient

if TYPE_CHECKING:
    from requests import Response


class Client(TestClient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.last_request: Optional['Response'] = None

    def request(self, *args, **kwargs) -> 'Response':
        r = super().request(*args, **kwargs)
        self.last_request = r
        return r

    def get_json(
        self, url: str, *, allow_redirects: bool = False, status: Optional[int] = 200, **kwargs
    ) -> Dict[str, Any]:
        r = self.get(url, allow_redirects=allow_redirects, **kwargs)
        if status:  # pragma: no cover
            assert r.status_code == status, r.text
        return r.json()

    def post_json(
        self, url: str, json: Any = None, *, allow_redirects: bool = False, status: Optional[int] = 200
    ) -> Dict[str, Any]:
        r = self.post(url, json=json, allow_redirects=allow_redirects)
        if status:
            assert r.status_code == status, r.text
        return r.json()
